@@ -121,7 +121,7 @@ crun_command_run (struct crun_global_arguments *global_args, int argc, char **ar
 
   if (bundle != NULL)
     if (chdir (bundle) < 0)
-      libcrun_fail_with_error (errno, "chdir '%s' failed", bundle);
+      libcrun_fail_with_error (errno, "chdir `%s` failed", bundle);
 
   container = libcrun_container_load_from_file ("config.json", err);
   if (container == NULL)
